%-----------------------------------------------------------------------------%
% vim: ft=prolog ts=4 sw=4 et wm=0 tw=0
%-----------------------------------------------------------------------------%
:-module(destruction, [ ':='/2,   op(700, xfx, :=),
                        ':=:'/2,  op(700, xfx, :=:),
                        iss/2,    op(700, xfx, iss),
                        '++'/1,   op(500, xf,  ++),
                        '--'/1,   op(500, xf,  --),
                        '+:+'/1,  op(500, xf,  +:+),
                        '-:-'/1,  op(500, xf,  -:-),
                        '++'/1,   op(500, fx,  ++),
                        '--'/1,   op(500, fx,  --),
                        '+:+'/1,  op(500, fx,  +:+),
                        '-:-'/1,  op(500, fx,  -:-),
                        defvar/1, op(400, fx,  defvar)  ]).


/** <module> Destructive variables (d-variables)
  *  
 * Aren't we all bored of NN is N+1? Why go through all this hassle?
 * Good news everyone: this module allows you to release your inner god of chaos
 * and use destructive assignment - in PROLOG!
 *
 * Of course, everything comes with a price; if you thought that
 * =/2 vs ==/2 vs =':='/2 vs is/2 is confusing then you might want to run...
 *
 *  @author      Thanos Tintinidis <thanosqr@gmail.com>
 *  @version     1.0
 *  @see         http://en.wikipedia.org/wiki/Alignment_%28Dungeons_%26_Dragons%29#Chaotic_Evil
 *  @deprecated
 *  @license     This program is free software. It comes without any warranty,
 *               to the extent permitted by applicable law. You can
 *               redistribute it and/or modify it under the terms of the Do
 *               What The Fuck You Want To Public License, Version 2, as
 *               published by Sam Hocevar.  See the file COPYING in the project
 *               root directory or consult http://sam.zoy.org/wtfpl/COPYING for
 *               more details.
 *  @tbd         Destructive assignments can never stop; therefore, this module
 *               is eternally in a state of flux
 */
  

%% defvar(Var:term) is det.
%
%  Since we are abandoning our high-level discipline, why not start declaring
%  our variables? defvar/1 allows you to declare one empty d-variable (but most
%  of the times it is pointless). Nevertheless, it demonstrates the basic
%  structure of our variables: v(_value) (where v stands for eVil and Variable)
%
%  ==
%  ?- defvar X.
%  X = v(_G666).
%  ==
defvar v(_).

%% :=(X:var, Y:var) is det.
%
%  ':='/2 will assign the value Y to X.  If Y contains instances of X, they
%  will be replaced by its value to avoid cyclic terms Note that X can be a
%  d-variable (empty or not) or a non-instantiated variable.
%
%  ==
%  ?- X := 42.
%  X = v(42).
%  ?- X := 42, print(X), X := 17, print(X).
%  4217
%  X = v(17).
%  ?- defvar(X), X := 42.
%  X = v(42).
%  ?-X:=1+1.
%  X = v(1+1)
%  ?- X:=1, X:=X+1.
%  X = v(1+1)
%  ?- X:=1, X := X+1+v(2)+v(3+v(32)).
%  X = v(1+1+v(2)+v(3+v(32))).
%  ==
%
%  @see ':=:'/2
X := Y :- (    var(X) -> X = v(Y)
          ;           ( replace_x(X,Y,YMX),
                        setarg(1,X,YMX) ) ).

%% :=:(X:var, Y:term) is det.
%
%  ':=:'/2 behaves similarly to ':='/2 but evaluates the expression Y before
%  assigning it to X.
%
%  ==
%  ?- X :=: 1.
%  X = v(1).
%  ?- X :=: 1+1.
%  X = v(2).
%  ?- X := 1+1.
%  X = v(1+1).
%  ?- X:=1, X :=: X+1.
%  X = v(2).
%  ?- X:=1, X :=: X+1+v(2)+v(3+v(32)).
%  X = v(39).
%  ==
%
%  @see ':='/2
X :=: Y :- YY iss Y,
           ( X = v(YY) -> true
           ;              setarg(1,X,YY) ).


%% iss(Y:var, X:term) is det.
%
%  iss/2 evaluates the expression X and assigns (non-destructively) the value
%  to Y.
%
%  ==
%  ?- Y iss 1+v(2)+v(3+v(32)).
%  Y = 38.
%  ?- v(Y) iss 1+v(2)+v(3+v(32)).
%  false.
%  ==

Y iss X :- term_to_atom(X,A),
           atom_chars(A,C),
           phrase(var_term(PC),C),
           atom_chars(PA,PC),
           atom_to_term(PA,T,[]),
           Y is T,!.

%% ++(X:var) is semidet.
%
%  Adds one to the value of a d-variable.  If the d-variable is empty, an
%  exception will be thrown.
%
%  Usage (covering ++/1, --/1, -:-/1, and +:+/1):
%  ==
%  ?- X := 1, X++ .
%  X = v(1+1).
%  ?- X := 1, ++X .
%  X = v(1+1).
%  ?- X := 1, +:+X.
%  X = v(2).
%  ?- X := 1, X++, X++ .
%  X = v(1+1+1).
%  ?- X := 1, X++, X++, +:+X.
%  X = v(4).
%  ==
%
%  @see +:+/1
%  @see --/1
%  @see -:-/1
(X ++) :- vcheck(X), X := X + 1.

%% +:+(X:var) is semidet.
%
%  Adds one to the value of a d-variable after evaluating it.  If the
%  d-variable is empty, an exception will be thrown.
%
%  @see --/1
%  @see +:+/1
%  @see -:-/1
(X +:+) :- vcheck(X), X :=: X + 1.

%% --(X:var) is semidet.
%
%  Subtracts one from the value of a d-variable.  If the d-variable is empty,
%  an exception will be thrown.
%
%  @see ++/1
%  @see +:+/1
%  @see -:-/1
(X --) :- vcheck(X), X := X - 1.
%
%% -:-(X:var) is semidet.
%
%  Subtracts one from the value of a d-variable after evaluating it.  If the
%  d-variable is empty, an exception will be thrown.
%
%  @see ++/1
%  @see --/1
%  @see +:+/1
(X -:-) :- vcheck(X), X :=: X - 1.

% Auxilliary predicates.  You are not meant to understand these.  They are
% evil.  Just trust us on this.

vcheck(v(X)):- ( var(X) -> throw('Arguments are not sufficiently instantiated')
               ;           true ).

var_term([]) -->     [].
var_term(PC) -->     [v,'('], { ! },
                     var_term(PC1),
                     [')'], { ! },
                     var_term(PC2),
                     { append(PC1, PC2, PC) }.
var_term([X|PC]) --> [X],
                     var_term(PC).

replace_x(X,Y,YMX):- term_chars(Y, YC),
                     term_chars(X, XC),
                     XC = ['v'| Value],
                     phrase(del_x(PC, {XC, Value}), YC),
                     term_chars(YMX, PC), !.

del_x([], _) --> [].
del_x(PC, {XC, Value}) --> XC, { ! },
                           del_x(PCT, {XC, Value}),
                           { append(Value, PCT, PC) }.
del_x([X|PC], XC) -->      [X],
                           del_x(PC, XC).

term_chars(T, C):- var(C),
                   term_to_atom(T, A),
                   atom_chars(A, C).

term_chars(T, C):- var(T),
                   atom_chars(A, C),
                   atom_to_term(A, T, []).
